import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import os
import warnings
from tqdm import tqdm
import argparse

# Make sure this matches your training file exactly
max_input_size = 10000

# Pixel art cat
def print_cat():
    cat_art = r"""
    /\_/\  
   ( o.o ) < Is this malware?
    > ^ <
    """
    print(cat_art)

def get_device():
    if torch.backends.mps.is_available():
        return torch.device("mps")
    elif torch.cuda.is_available():
        return torch.device("cuda")
    else:
        return torch.device("cpu")

class MalwareClassificationModel(nn.Module):
    def __init__(self):
        super(MalwareClassificationModel, self).__init__()
        self.fc1 = nn.Linear(max_input_size, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, 64)
        self.fc4 = nn.Linear(64, 2)
        self.dropout = nn.Dropout(0.5)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        x = F.relu(self.fc2(x))
        x = self.dropout(x)
        x = F.relu(self.fc3(x))
        x = self.fc4(x)
        return x

def read_binary(file_path):
    with open(file_path, 'rb') as f:
        binary_data = f.read()

    binary_array = np.frombuffer(binary_data, dtype=np.uint8)
    binary_array = np.pad(binary_array, (0, max(0, max_input_size - len(binary_array))), 'constant')[:max_input_size]
    return torch.tensor(binary_array, dtype=torch.float32)

def load_model(model_path):
    device = get_device()
    model = MalwareClassificationModel().to(device)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        state_dict = torch.load(model_path, map_location=device)
        model.load_state_dict(state_dict)
    model.eval()
    return model, device

def classify_binary(input_tensor, model, device):
    input_tensor = input_tensor.unsqueeze(0).to(device)

    with torch.no_grad():
        outputs = model(input_tensor)
        probabilities = torch.softmax(outputs, dim=1)
        predicted_class = outputs.argmax(dim=1).item()
        confidence = probabilities[0][predicted_class].item()

    prediction = "Malicious" if predicted_class == 1 else "Benign"
    return prediction, confidence

def classify_file(file_path, model, device):
    input_tensor = read_binary(file_path).to(device)
    prediction, confidence = classify_binary(input_tensor, model, device)

    print("\n" + "=" * 60)
    print(f"Scanning File: {file_path}")
    print("=" * 60)
    print(f"Prediction      : {prediction}")
    print(f"Confidence      : {confidence:.4f}")
    print(f"Raw Outputs     : {model(input_tensor.unsqueeze(0)).squeeze().tolist()}")
    print("=" * 60 + "\n")

def classify_directory(directory_path, model, device):
    files = [f for f in os.listdir(directory_path) if os.path.isfile(os.path.join(directory_path, f))]

    results = []
    for filename in tqdm(files, desc="Classifying files", unit="file"):
        file_path = os.path.join(directory_path, filename)
        input_tensor = read_binary(file_path)
        prediction, confidence = classify_binary(input_tensor, model, device)
        results.append((filename, prediction, confidence))

        print("\n" + "=" * 60)
        print(f"File: {filename}")
        print("=" * 60)
        print(f"Prediction      : {prediction}")
        print(f"Confidence      : {confidence:.4f}")
        print("=" * 60 + "\n")

    return results

def main():
    print_cat()
    parser = argparse.ArgumentParser(description="Malware classification tool with a friendly interface")
    parser.add_argument("--file", type=str, help="Path to a single file to scan")
    parser.add_argument("--directory", type=str, help="Path to a directory of files to scan")
    parser.add_argument("--model", type=str, help="Path to the trained model file (.pth)")

    args = parser.parse_args()

    # Prompt for model path if not provided
    model_path = args.model
    while not model_path or not os.path.isfile(model_path):
        model_path = input("Please enter the path to your trained model file (.pth): ")
        if not os.path.isfile(model_path):
            print("Invalid file path. Please try again.")

    # Load the model
    model, device = load_model(model_path)

    if args.file and args.directory:
        print("Please specify either a file or a directory, not both.")
        return

    if args.file:
        if os.path.isfile(args.file):
            print(f"Scanning file: {args.file}")
            classify_file(args.file, model, device)
        else:
            print(f"File {args.file} does not exist.")
    elif args.directory:
        if os.path.isdir(args.directory):
            print(f"Scanning directory: {args.directory}")
            results = classify_directory(args.directory, model, device)
            # Print summary
            print("\nClassification Summary:")
            benign_count = sum(1 for _, pred, _ in results if pred == "Benign")
            malicious_count = sum(1 for _, pred, _ in results if pred == "Malicious")

            print(f"Total files    : {len(results)}")
            print(f"Benign         : {benign_count}")
            print(f"Malicious      : {malicious_count}")
            print(f"Benign ratio   : {benign_count / len(results):.2%}")
            print(f"Malicious ratio: {malicious_count / len(results):.2%}")
        else:
            print(f"Directory {args.directory} does not exist.")
    else:
        print("Please specify either a file or a directory to scan.")

if __name__ == "__main__":
    main()